/*
Copyright (c) 2008,  Felipe Lessa

Developed by Felipe Lessa (felipe.lessa@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining 
a copy of this software and associated documentation files (the 
"Software"), to deal in the Software without restriction, including 
without limitation the rights to use, copy, modify, merge, publish, 
distribute, sublicense, and/or sell copies of the Software, and to 
permit persons to whom the Software is furnished to do so, subject 
to the following conditions:

The above copyright notice and this permission notice shall be 
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR 
ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF 
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION 
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

(MIT license, http://www.opensource.org/licenses/mit-license.html)
 */
package lars.tools;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.util.Observable;

import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JTextField;

import lars.ProgramStatement;
import lars.loongarch.hardware.AccessNotice;
import lars.loongarch.hardware.AddressErrorException;
import lars.loongarch.hardware.Memory;
import lars.loongarch.hardware.MemoryAccessNotice;
import lars.loongarch.instructions.BasicInstruction;
import lars.loongarch.instructions.BasicInstructionFormat;

/**
 * 
 * Instruction counter tool. Can be used to know how many instructions
 * were executed to complete a given program. 
 * 
 * Code slightly based on MemoryReferenceVisualization.
 * 
 * @author Felipe Lessa <felipe.lessa@gmail.com>
 *
 */
//@SuppressWarnings("serial")
public class InstructionCounter extends AbstractLarsToolAndApplication {
    private static String name    = "Instruction Counter";
    private static String version = "Version 1.0 (Felipe Lessa)";
    private static String heading = "Counting the number of instructions executed";



    /**
     * Number of instructions executed until now.
     */
    protected int counter = 0;
    private JTextField counterField;

	/**
	 * Number of instructions of type
	 */
	class InstructionTypeCounter{
		protected int counterType=0;
		//private int type;
		private String typeName;
		private JTextField counterTypeField;
		private JProgressBar progressBarType;

		InstructionTypeCounter(String typeName){
			this.typeName=typeName;
		}
	}

	private InstructionTypeCounter[] istc=new InstructionTypeCounter[]{
			new InstructionTypeCounter("TWO_R_TYPE"),
			new InstructionTypeCounter("THREE_R_TYPE"),
			new InstructionTypeCounter("FOUR_R_TYPE"),
			new InstructionTypeCounter("TWO_R_I8_TYPE"),
			new InstructionTypeCounter("TWO_R_I12_TYPE"),
			new InstructionTypeCounter("TWO_R_I14_TYPE"),
			new InstructionTypeCounter("TWO_R_I16_TYPE"),
			new InstructionTypeCounter("JUMP_TYPE"),
			new InstructionTypeCounter("ONE_R_I21_TYPE"),
			new InstructionTypeCounter("ONE_R_I20_TYPE"),
			new InstructionTypeCounter("I26_TYPE"),
			new InstructionTypeCounter("SYSCALL"),
	};


    /**
     * The last address we saw. We ignore it because the only way for a
     * program to execute twice the same instruction is to enter an infinite
     * loop, which is not insteresting in the POV of counting instructions.
     */
    protected int lastAddress = -1;
    
   	/**
   	 * Simple constructor, likely used to run a stand-alone memory reference visualizer.
   	 * @param title String containing title for title bar
   	 * @param heading String containing text for heading shown in upper part of window.
   	 */
    public InstructionCounter(String title, String heading) {
    	super(title,heading);
    }
    
    /**
     * Simple construction, likely used by the LARS Tools menu mechanism.
     */
    public InstructionCounter() {
    	super(name + ", " + version, heading);
    }

//	@Override
	public String getName() {
		return name;
	}



//	@Override
	protected JComponent buildMainDisplayArea() {
		// Create everything
		JPanel panel = new JPanel(new GridBagLayout());

		counterField = new JTextField("0", 10);
		counterField.setEditable(false);

		for (int i = 0; i < 12; i++) {
			istc[i].counterTypeField=new JTextField("0",10);
			istc[i].counterTypeField.setEditable(false);
			istc[i].progressBarType=new JProgressBar(JProgressBar.HORIZONTAL);
			istc[i].progressBarType.setStringPainted(true);
		}

		
		// Add them to the panel
		
		// Fields
		GridBagConstraints c = new GridBagConstraints();
		c.anchor = GridBagConstraints.LINE_START;
		c.gridheight = c.gridwidth = 1;
		c.gridx = 3;
		c.gridy = 1;
		c.insets = new Insets(0, 0, 17, 0);
		panel.add(counterField, c);

		c.insets = new Insets(0, 0, 0, 0);

		for (int i = 0; i < 12; i++) {
			c.gridy++;
			panel.add(istc[i].counterTypeField,c);
		}

		
		// Labels
		c.anchor = GridBagConstraints.LINE_END;
		c.gridx = 1;
		c.gridwidth = 2;
		c.gridy = 1;
		c.insets = new Insets(0, 0, 17, 0);
		panel.add(new JLabel("Instructions so far: "), c);

		c.insets = new Insets(0, 0, 0, 0);
		c.gridx = 2;
		c.gridwidth = 1;

		for (int i = 0; i < 12; i++) {
			c.gridy++;
			panel.add(new JLabel(istc[i].typeName+":"),c);
		}
		
		// Progress bars
		c.insets = new Insets(3, 3, 3, 3);
		c.gridx = 4;
		c.gridy = 2;

		for (int i = 0; i < 12; i++) {
			panel.add(istc[i].progressBarType,c);
			c.gridy++;
		}

		
		return panel;
	}
	
//	@Override
	protected void addAsObserver() {
		addAsObserver(Memory.textBaseAddress, Memory.textLimitAddress);
	}

//	@Override
	protected void processLoongArchUpdate(Observable resource, AccessNotice notice) {
		if (!notice.accessIsFromLoongArch()) return;
		if (notice.getAccessType() != AccessNotice.READ) return;
		MemoryAccessNotice m = (MemoryAccessNotice) notice;
		int a = m.getAddress();
		if (a == lastAddress) return;
		lastAddress = a;
		counter++;
		try {
			ProgramStatement stmt = Memory.getInstance().getStatement(a);
			BasicInstruction instr = (BasicInstruction) stmt.getInstruction();
			BasicInstructionFormat format = instr.getInstructionFormat();

			if(format==BasicInstructionFormat.TWO_R_TYPE)
				istc[0].counterType++;
			else if(format==BasicInstructionFormat.THREE_R_TYPE)
				istc[1].counterType++;
			else if(format==BasicInstructionFormat.FOUR_R_TYPE)
				istc[2].counterType++;
			else if(format==BasicInstructionFormat.TWO_R_I8_TYPE)
				istc[3].counterType++;
			else if(format==BasicInstructionFormat.TWO_R_I12_TYPE)
				istc[4].counterType++;
			else if(format==BasicInstructionFormat.TWO_R_I14_TYPE)
				istc[5].counterType++;
			else if(format==BasicInstructionFormat.TWO_R_I16_TYPE||format==BasicInstructionFormat.TWO_R_I16_BRANCH_TYPE)
				istc[6].counterType++;
			else if(format==BasicInstructionFormat.JUMP_TYPE)
				istc[7].counterType++;
			else if(format==BasicInstructionFormat.ONE_R_I20_TYPE)
				istc[8].counterType++;
			else if(format==BasicInstructionFormat.ONE_R_I21_TYPE)
				istc[9].counterType++;
			else if(format==BasicInstructionFormat.I26_TYPE)
				istc[10].counterType++;
			else if(format==BasicInstructionFormat.SYSCALL)
				istc[11].counterType++;
			// if (format == BasicInstructionFormat.R_FORMAT)
			// 	counterR++;
			// else if (format == BasicInstructionFormat.I_FORMAT
			// 		|| format == BasicInstructionFormat.I_BRANCH_FORMAT)
			// 	counterI++;
			// else if (format == BasicInstructionFormat.J_FORMAT)
			// 	counterJ++;
		} catch (AddressErrorException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		updateDisplay();
	}
	
//	@Override
	protected void initializePreGUI() {
		counter =0;
		for (int i = 0; i < 12; i++) {
			istc[i].counterType=0;
		}

		lastAddress = -1;
	}
	
// @Override
	protected void reset() {
		counter=0;
		for (int i = 0; i < 12; i++) {
			istc[i].counterType=0;
		}
		lastAddress = -1;		
		updateDisplay();
	}
	
//	@Override
	protected void updateDisplay() {
		counterField.setText(String.valueOf(counter));

		for (int i = 0; i < 12; i++) {
			istc[i].counterTypeField.setText(String.valueOf(istc[i].counterType));
			istc[i].progressBarType.setMaximum((counter));
			istc[i].progressBarType.setValue(istc[i].counterType);
		}


		
		if (counter == 0) {

			for (int i = 0; i < 12; i++) {
				istc[i].progressBarType.setString("0%");
			}
		} else {
			for (int i = 0; i < 12; i++) {
				istc[i].progressBarType.setString(((istc[i].counterType * 100)/counter+ "%"));
			}
		}
	}
}
